function varargout = illrfittedqi(cfg)
% Fitted Qs-iteration with local linear regression and interlaced sample generation.
%   APPROX = FITTEDQI(CFG)              - in 'run' mode
%   [HIST, FIGH] = FITTEDQI(CFG)        - in 'replay' mode
%   FIGH = FITTEDQI(CFG)                - in 'sol' or 'evol' modes
% An implementation of fitted Q-iteration with local linear regression
% approximation, and interlaced, repeated generation of samples using intermediate solutions.
% Note the algorithm assumes a deterministic system
%
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see str2cfg
% 
% Outputs:
%       APPROX      - final Q-function & policy approximator
%                   (non-parametric)
%       HIST        - the replay history (trajectories)
%       FIGH        - handles to the figures created

if nargin < 1, cfg = struct(); end;

% ==== DECLARE CONFIGURATION DEFAULTS ====
% function config
CFG.run = 0;                        % run learning
CFG.replay = 0;                     % replay learned policy 
CFG.sol = 0;                        % plot solution
CFG.evol = 0;                       % plot evolution (once every evol iterations)
CFG.datafile = 'illrfittedqidata';   % save data to file
CFG.datadir = [];                   % data dir
% main algorithm config
CFG.problem = '';                   % what problem to solve
CFG.gamma = [];                     % discount factor
CFG.U = [];                         % discrete action space (cell array of grids, or flat)
% LLR config
CFG.knn = 10;                       % # of samples
CFG.S = 'auto';                     % distance scaling: [] = no scaling, dim(x)-vector = diagonal scaling
                                    % 'auto' choose to scale state variables equally
CFG.wfun = 'unif';                  % weighting fun (kernel): [tri]'cube', 'quad', 'unif', 'triang'
CFG.knnmethod = 'jit';              % method for knn search; only 'jit' (Matlab JIT-optimized) supported
CFG.linsolvemethod = 'pinv';        % only 'pinv' supported (cannot use mldivide with precomputation)
CFG.extrap = 'sat';                 % extrapolation handling: 'sat' = saturate, 'none' = leave alone
CFG.duplicates = 'remove';          % how to handle duplicates: 'remove', 'keep'
% samples, iterations etc.
CFG.maxtau = 10;                    % # of metaiterations to run
CFG.maxiter = 50;                   % max # of inner iterations to run per metaiteration
CFG.Nstau = 1000;                   % # of new samples to generate at each metaiter
CFG.maxsteps = 100;                 % maximum length of each trajectory (preferably divisor of Nstau)
CFG.X0 = [];                        % set of initial states from which to simulate
CFG.h0 = 'rand';                    % initial policy, or 'rand' for uniform random
CFG.h0args = {};                    % initial policy arguments
CFG.explor = .2;                    % exploration probability
% convergence checks
CFG.eps = 0.1;                      % convergence threshold
CFG.maxdelta = 1e15;                % divergence threshold
CFG.checkosc = 1;                   % whether to check for oscillations
CFG.maxper = 10;                    % osc: maximum period to check (# iter) -- >= 2
CFG.nper = 1;                       % osc: number of periods to check
CFG.epsper = 0.1;                   % osc: maximum abs diff between two par vectors in diff periods
% replay config
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay
% stats config
CFG.storehistory = 1;               % solution history: =1: at every meta-iteration (approx only); =2: also at inner iterations
% display config
CFG.verb = 3;                       % verbosity: the higher, the more detailed the messages displayed
CFG.iterdisp = 10;                  % feedback after every iterdisp iterations
CFG.tausave = 1;                    % save after each tausave META-iterations
CFG.sampledisp = .05;               % feedback once this % of samples processed
CFG.silent = 0;                     % suppress all output
% figure config
CFG = setfigprop(CFG, 'addfields');
CFG.slice = [];                     % slice of solution to plot
CFG.miscinfo = '';                  % field for miscellaneous information

% Early defaults (initialized before calling problem defaults)
ECFG.model_params = {};             % parameters for problem calling in 'model' mode
ECFG.fittedqi_params = {};          % parameters for problem calling in 'fittedqi' mode

% List of fields that cannot change in-between recalls of the function
% Currently not used since we do not support a "resume" mode
KEEPFIELDS = {'problem', 'gamma'};

% ==== PARSE AND PROCESS CONFIG ====
cfg = parseconfig(cfg, CFG, ECFG, 'fittedqi');
% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% If running, check presence of data file, warn on possible overwrite
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '.mat] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;       
% determine if initialization needed; if not, need to load: check if datafile exists
cfg.init = ~(cfg.replay || cfg.sol || cfg.evol);
if ~cfg.init && ~exist([cfg.datafile '.mat'], 'file'),
    error(['File [' cfg.datafile '.mat] does not exist. Terminating.']);
end;

% ==== IF NOT INITIALIZING: NEED TO LOAD DATA ====
if ~cfg.init,        % load data file, making sure that cfg and KEEPFIELDS is not overwritten
    % optimize the loading time: only load large variables when needed
    dfv = who('-file', cfg.datafile);
    % REMARK remove here any large variables unnecessary for the current mode 
    if ~cfg.evol, dfv = rmstring(dfv, 'Qsh'); end;
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile, dfv{:});
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['iLLRFittedQI: data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

% Some error checking
if ~strcmp(cfg.linsolvemethod, 'pinv'), 
    error('Linear solving method [%s] unknown (only [pinv] unsupported).', cfg.linsolvemethod); 
end;
if ~strcmp(cfg.knnmethod, 'jit'), 
    error('kNN method [%s] unsupported.', cfg.knnmethod); 
end;
if ~any(strcmp(cfg.extrap, {'sat', 'none'})), 
    error('Extrapolation handling method [%s] unknown.', cfg.extrap); 
end;

% Echo config
dispx('Interlaced-LLRFittedQI will run with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');

% ==== IF INITIALIZING: CREATE MODEL, ETC. ====
if cfg.init,
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    if isempty(cfg.U) && ~iscell(cfg.U), 
        error('Predefined, discrete action space required as a cell array of grids.'); 
    end;
end;

% ==== RUN FITTED QI ====
if cfg.run,
       
    % Initialize if not resuming
    if cfg.init,
        % shorthand variables
        p = model.p; q = model.q;
        knn = cfg.knn;
        switch cfg.extrap,
            case 'none',    satflag = 0;
            case 'sat',     satflag = 1;
        end;

        % flatten action space
        U = flat(cfg.U); 
        M = size(U, 2);     % number of discrete actions

        % distance scaling
        if isempty(cfg.S),
            cfg.S = ones(p, 1);
        elseif ischar(cfg.S) && strcmp(cfg.S, 'auto'),
            if isfield(model, 'minx'),      cfg.S = 1 ./ (model.maxx - model.minx);
            else                            cfg.S = 1 ./ model.maxx;    % assumed symmetrical
            end;
            % normalize distance weighting
            % cfg.S = cfg.S ./ sum(cfg.S);
        elseif length(cfg.S) ~= model.p,
            error('Scaling vector S is incorrectly sized');
        end;
        
        % samples
        Ns = 0;
        Xs = nan(p, 0); Us = nan(q, 0); Xps = nan(p, 0); Rs = nan(1, 0);
        Qs = nan(1, 0);
        
        % history
        deltah = nan(cfg.maxtau, cfg.maxiter+1); 
        iter = nan(cfg.maxtau, 1);
        if cfg.storehistory >= 2 || cfg.checkosc, Qsh = cell(cfg.maxtau, cfg.maxiter); end;
        if cfg.storehistory >= 1, approxh = cell(cfg.maxtau, 1); end;
        
        % convergence flags
        conv = NaN(cfg.maxtau, 1);
        
        % execution time statistics
        timestat = struct;
        timestat.run = 0;
        
        tau = 1;            % init meta-iteration
    end;        % initialization IF
        
    dispx('Performing iLLRFittedQI...', cfg.verb, 0);
    
    % Main Q-iteration loop
    while tau <= cfg.maxtau && ~(tau > 1 && conv(tau-1) == -1), % stop if divergence detected
        
        dispx(sprintf('Running meta-iter tau=%d... ', tau), cfg.verb, 1);
        timestat.tau_run(tau) = 0;
        
        % Generate (new) samples
        dispx(sprintf('  generating %d samples... ', cfg.Nstau), cfg.verb, 2);
        tmark = cputime;
        % constant config
        scfg = struct;
        scfg.method = 'trajdisc';
        scfg.ugrids = cfg.U;        % note we use the cell-array form on the config
        scfg.N = cfg.Nstau; 
        scfg.traj_K = cfg.maxsteps;
        scfg.traj_X0 = cfg.X0; 
        scfg.explor = cfg.explor;
        % policy
        if tau > 1,     % regular iteration -- use current approx
            scfg.policy = @(x, t) approx.h(approx, Qs, x);
        else            % first iteration -- use configured initial policy
            if strcmp(cfg.h0, 'rand'),
                scfg.policy = @hrand; scfg.policyargs = {U, 1};    % 1 = discrete output
            else
                scfg.policy = cfg.h0; scfg.policyargs = cfg.h0args;
            end;
        end;
        % 3rd argument = dummy approximator to transmit the discrete action sapce
        % note we provide the vectorized (non-flat) action space
        samples = generate_samples(model, scfg);

        % pick up samples
        Ns = Ns + samples.N; 
        Xs = [Xs samples.X]; Us = [Us samples.U]; 
        Xps = [Xps samples.Xp]; Rs = [Rs samples.R]; 

        % init new Q-samples
        % Qs = zeros(Ns, 1);    % can init at zeros, but pbly better to exploit existing values
        Qsold = Qs; 
        Qs = [Qs; nan(samples.N, 1)];
        switch cfg.duplicates,
            case 'keep',        % do nothing
            case 'remove',      
                dispx([8 'removing duplicates... '], cfg.verb, 2);
                % remove duplicates; 
                % use 1e-10 to account for numerical errors in simulating transitions etc.
                %
                % note we assume here that a given (x, u) will always lead to the same (xp, rp)
                % (i.e., deterministic dynamics and reward function)
                %
                % the aggregate Q-value is the mean of (existing) Q-values
                % Q-values of new samples are nan so they are not taken into account in the mean
                % the aggregate sample is exactly one of the original samples (no averaging)
                [XUXR, Qs] = removeduplicates([Xs' Us' Xps' Rs'], Qs, 1e-10);
                % count and retrieve unique samples
                Ns = size(XUXR, 1);
                Xs = XUXR(:, 1:p)'; Us = XUXR(:, p+1:p+q)'; 
                Xps = XUXR(:, p+q+1:p+q+p)'; Rs = XUXR(:, p+q+p+1)';
                % note even after removal, completely new samples (for which duplicates with
                % Q-values did not exist) will still have uninitialized (nan) Q-values
        end;
        % record time taken to generate samples and possibly remove duplicates
        timestat.tau_generatesamples(tau) = cputime - tmark;
        timestat.tau_run(tau) = timestat.tau_run(tau) + timestat.tau_generatesamples(tau);        
        dispx([8 sprintf('done; now using %d samples', Ns)], cfg.verb, 2);       % "generating samples" done
        
        tmark = cputime;
        if tau == 1,
            % make Qvalues zero
            Qs(isnan(Qs)) = 0;
        else
            % we ran at least one iteration, use the current approx to guess initial values
            for is = find(isnan(Qs)), Qs(is) = approx.q(approx, Qsold, Xs(:, is), Us(:, is)); end;
        end;
        timestat.tau_computenewQs(tau) = cputime - tmark;
        timestat.tau_run(tau) = timestat.tau_run(tau) + timestat.tau_computenewQs(tau);

        % clear no longer needed, memory-intensive variables
        clear samples scfg Qsold

        % Identify portion of sample data corresponding to each discrete action
        Is = cell(M, 1); for j = 1:M, Is{j} = findflat(U(:, j), Us); end;

        % Pre-computation of k nearest neighbors and partial inverses for
        % each sample x discrete action
        % Note we need to do this every metaiter for all samples, as neighbors may have changed
        
        % Phase 1) search for k nearest neighbors
        dispx('  pre-searching kNNs...', cfg.verb, 2);
        tmark = cputime;            % notice console output below COUNTED IN
        INDLLR = nan(knn, Ns, M);   % indices of NNs
        WLLR = nan(knn, Ns, M);     % initially distances to NNs; converted to weights at next step
        % this code runs with discrete-action granularity, for all state samples
        dispx([8 8 8 8 sprintf(' - disc actions processed: %2d', 0)], cfg.verb, 3);
        for j = 1:M,
            % find indices & distances for all Xps for discrete action j (notice transposes)
            [ind, dist] = knnsearchx(Xps', Xs(:, Is{j})', knn, cfg.S);
            % transform to global indices
            ind(:) = Is{j}(ind(:));
            % save indices and distances to arrays (notice transpose)
            INDLLR(:, :, j) = ind';
            WLLR(:, :, j) = dist';
            if cfg.verb >= 3, disp([8 8 8 sprintf('%2d', j)]); end;     % display progress inline
        end;
        timestat.tau_findknn(tau) = cputime - tmark;
        timestat.tau_run(tau) = timestat.tau_run(tau) + timestat.tau_findknn(tau);
        dispx([8 ' done.'], cfg.verb, 2);
        
        % Phase 2) precompute partial inverses
        dispx('  pre-computing pinv...', cfg.verb, 2);
        dispx([8 8 8 8 sprintf(' - samples processed: %3d%%', 0)], cfg.verb, 3);
        tmark = cputime;
        XLLRpinv = nan(p+1, knn, Ns, M);
        xllr = zeros(knn, p);   
        for is = 1:Ns,
            for j = 1:M,
                % transform the distances to the current sample into actual
                % weights, using the weight function
                WLLR(:, is, j) = llrweights(cfg.wfun, WLLR(:, is, j), WLLR(knn, is, j));
                % for each kNN: subtract query sample and multiply by weight
                % notice data points are on the rows of the LHS (xllr)
                % (rely on JIT to optimize this for, rather than using repmat)
                for ik = 1:knn,
                    xllr(ik, :) = WLLR(ik, is, j) .* (Xs(:, INDLLR(ik, is, j)) - Xps(:, is));
                end;
                % last column (weighted) biases, i.e., 1 * weights, i.e., weights
                XLLRpinv(:, :, is, j) = pinv([xllr WLLR(:, is, j)]);
            end;
            if cfg.verb >= 3 && mod(is/Ns, cfg.sampledisp) <= 1/Ns,    % display progress inline
                disp([8 8 8 8 8 sprintf('%3d%%', fix(is/Ns*100))]);
            end;
        end; 
        timestat.tau_computepinv(tau) = cputime - tmark;
        timestat.tau_run(tau) = timestat.tau_run(tau) + timestat.tau_computepinv(tau);
        dispx([8 ' done.'], cfg.verb, 2);        

        % store initial Q-values
        if cfg.storehistory >= 2 || cfg.checkosc, Qsh{tau, 1} = Qs; end;
        
        % run inner loop
        conv(tau) = 0;               % reset convergence flag
        k = 1;                  % init inner iteration index
        Qxp = zeros(M, 1);
        while k <= cfg.maxiter && ~conv(tau),

            dispx(sprintf('    inner-iter k=%d, samples processed: %3d%%', k, 0), cfg.verb, 4);
            timestat.iter_run(tau, k) = 0;

            tmark = cputime;
            Qsold = Qs; 
            for is = 1:Ns,
                % find approximate Q-values of all the discrete actions
                for j = 1:M,
                    % pick up Q samples for NNs 
                    qllr = Qsold(INDLLR(:, is, j));
                    % solve linear system to find local regressor; note Q samples are weighted;
                    % also note query point is subtracted from data, so we only need the bias term;
                    % the entire solution would be: 
                    % betallr = XLLRpinv(:, :, is, j) * (qllr .* WLLR(:, is, j));
                    biasbetallr = XLLRpinv(p+1, :, is, j) * (qllr .* WLLR(:, is, j));
                    % compute regression output for query point;
                    if satflag, Qxp(j) = max(min(qllr), min(biasbetallr, max(qllr))); 
                    else        Qxp(j) = biasbetallr;
                    end;
                end;
                Qs(is) = Rs(is) + cfg.gamma .* max(Qxp);
                % display progress inline -- note this IS counted in
                % runtime -- too much headache to remove it
                if cfg.verb >= 4 && mod(is/Ns, cfg.sampledisp) <= 1/Ns,
                    disp([8 8 8 8 8 sprintf('%3d%%', fix(is/Ns*100))]);
                end;
            end;

            % update runtime (statistics and user feedback below are not counted)
            timestat.iter_updateQ(tau, k) = cputime - tmark;
            timestat.iter_run(tau, k) = timestat.iter_run(tau, k) + timestat.iter_updateQ(tau, k);

            % save Q samples on history if configured
            if cfg.storehistory >= 2 || cfg.checkosc, Qsh{tau, k+1} = Qs; end;

            % perform convergence, divergence, oscillation tests
            tmark = cputime;
            deltah(tau, k) = max(abs(Qs - Qsold));
            if deltah(tau, k) <= cfg.eps,            
                conv(tau) = 1;
            elseif deltah(tau, k) > cfg.maxdelta,    
                conv(tau) = -1;
            elseif cfg.checkosc,    % check for oscillations
                per = 2; osc = 0;   % don't test for period = 1, that's taken care of by the simple convergence test
                % try periods up to maxper, as long as oscillation wasn't detected yet
                % and provided that the number of iterations allows for it
                while ~osc && (per <= cfg.maxper) && (per * (cfg.nper+1) + 1 < k+1)
                    osc = 1;                % start by assuming oscillating; tests below might falsify that
                    for i = 0:per-1,        % i loops through the offsets of the iters that have to be equal
                        for j = 1:cfg.nper, % make sure oscillation happens for at least nper periods
                            osc = osc & all(abs(Qsh{tau, k+1-i} - Qsh{tau, k+1-j*per-i}) <= cfg.epsper);
                            if ~osc, break; end;    % don't test longer periods
                        end;
                        if ~osc, break; end;        % don't test other offets
                    end;
                    per = per + 1;
                end;
                per = per - 1;                      % revert the last increment
                if osc, conv(tau) = 2; end;    % flag oscillation
                % clean up any unnecessary history
                if cfg.storehistory < 2 && k > cfg.maxper * (cfg.nper+1), Qsh{tau, k - cfg.maxper * (cfg.nper+1)} = []; end;
            end;
            timestat.iter_convtest(tau, k) = cputime - tmark;
            timestat.iter_run(tau, k) = timestat.iter_run(tau, k) + timestat.iter_convtest(tau, k);

            % finalize high-verbosity iteration display 
            dispx(sprintf([8 ' done, delta=%g.'], deltah(tau, k)), cfg.verb, 4);
            % visual feedback of algorithm progress
            if ~mod(k, cfg.iterdisp) && cfg.verb < 4, % if >= 4, taken care of by the code above
                dispx(sprintf('    inner-iter k=%d done, delta=%g.', k, deltah(tau, k)), cfg.verb, 3);
            end;

            timestat.tau_run(tau) = timestat.tau_run(tau) + timestat.iter_run(tau, k);
            k = k + 1;
        end;        % WHILE allowed more iterations
        
        iter(tau) = k - 1;
        
        % termination message 
        if cfg.verb <= 1,   pref = [8 8 8 8 ' done: '];   % we go on the same line as the metaiter start message
        else                pref = sprintf('Meta-iter tau=%d done: ', tau);
        end;
        if conv(tau) == 1,      dispx([pref 'convergence detected after ' num2str(k-1) ' iter.'], cfg.verb, 1);
        elseif conv(tau) == 2,  dispx([pref 'solution oscillates with period ' num2str(per)], cfg.verb, 1);
        elseif conv(tau) == -1, dispx([pref 'divergence detected!'], cfg.verb, 1);
        else                    dispx([pref 'maxiter=' num2str(cfg.maxiter) ' exhausted.'], cfg.verb, 1);
        end;

        % create approximator given current data
        acfg = struct;
        acfg.U = cfg.U;
        acfg.knnmethod = cfg.knnmethod; acfg.knn = cfg.knn; 
        acfg.S = cfg.S;
        acfg.wfun = cfg.wfun;
        acfg.linsolvemethod = cfg.linsolvemethod;
        acfg.extrap = cfg.extrap;
        acfg.Xs = Xs; acfg.Is = Is; acfg.Qs = Qs;
        approx = llrdisc(model, acfg);
        % store on history if configured
        if cfg.storehistory >= 1, approxh{tau} = approx; end;
        
        % data backup
        if ~mod(tau, cfg.tausave),
            save(cfg.datafile);
            dispx(sprintf('Data at meta-iter tau=%d saved to [%s].', tau, cfg.datafile), cfg.verb, 1);
        end;
        
        timestat.run = timestat.run + timestat.tau_run(tau);
        tau = tau + 1;  % increment meta-iteration
    end;

    % we are not checking for convergence etc. across meta-iterations
    if conv(tau-1) == -1,   dispx('Divergence detected! Algorithm interrupted.', cfg.verb, 0);
    else                    dispx(['maxtau=' num2str(cfg.maxtau) ' exhausted. Algorithm stopped.'], cfg.verb, 0);
    end;

    % use this line to cleanup history, if only used to check for oscillation
    % for now, keep it in case we need to resume later
    % if cfg.checkosc && cfg.storehistory < 2, clear Qsh; end;
end;


% Options below might create figures, init figure handles array
figh = [];
if isempty(cfg.savefig), cfg.savefig = cfg.datafile; end;

% ==== REPLAY POLICY -- Including intermediate plots and trajectory saving due to high comp cost ====
if cfg.replay,

    % initial state
    if ~isempty(cfg.x0), 
        if ischar(cfg.x0) && ~isempty(cfg.problem),
            % use the model function to get a string-named initial state
            x0 = feval(cfg.problem, 'x0', cfg.x0);
        else
            x0 = cfg.x0(:);         % specified initial state
        end;
    else    % try getting default initial state
        try         x0 = feval(cfg.problem, 'x0');
        catch       x0 = zeros(model.p, 1);    % zeros
        end;
    end;         
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    % init history
    t = 0 : model.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(model.p, length(t)); x(:, 1) = x0;
    u = zeros(model.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    th = NaN;   % time elapsed while computing action for one state
    for k = 1:Ns,
        dispx(sprintf('k=%d/%d. th=%f', k, Ns, th), cfg.verb, 5);
        u(:, k) = approx.h(approx, Qs, x(:, k));
        [x(:, k+1) r(k+1) terminal] = feval(model.fun, model, x(:, k), u(:, k));
        if terminal, Ns = k; break; end;
    end;
 
    % create & optionally save history
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);
    hist.R = discreturn(cfg, hist.r, Ns, terminal);
    % plot history & optionally save figures
    if ~cfg.silent,
        if isfield(model, 'plotfun'),
            mpfigh = feval(model.plotfun, hist);
            if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
        end;
        % if no plot function or plot function did not create a figure
        if ~isfield(model, 'plotfun') || isempty(mpfigh),
            figh(end+1) = plothistory(hist);
        end;
        setfigprop(cfg);
        % save if requested
        saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;        % IF replay

% ==== STATISTICS ====
if (cfg.evol || cfg.sol) && ~cfg.silent,
    % grayscale styles
    sty.cm = gray(128); sty.cm = sty.cm(33:end-5, :);  % w/o strong blacks and whites
    % labels
    labels.iter = 'Iteration'; 
    labels.h = 'h(x)'; labels.V = 'V(x)'; 
    labels.delta = {'delta'}; 
    figh = [];
end;

if cfg.sol && ~cfg.silent,
    figh(end+1) = figurex([1100 450]);
    K = find(~isnan(deltah), 1, 'last') - 1;
    subplot(3, 2, [1 3]); cla;
    pcfg = struct; pcfg.npoints = 30; pcfg.slice = cfg.slice;
    approx.plotv(approx, Qs, pcfg); title(labels.V);
    subplot(3, 2, [2 4 6]); cla;
    pcfg = struct; pcfg.npoints = 70; pcfg.slice = cfg.slice;
    approx.ploth(approx, Qs, pcfg); title(labels.h);
    subplot(3, 2, 5); cla;
    semilogy(1:K, deltah(1:K), 'k'); grid on; xlabel(labels.iter); ylabel(labels.delta{:});
    setfigprop(cfg);
    % save if requested
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;

if cfg.evol && ~cfg.silent,
    % TODO
    warning('illrfittedqi: evol mode not implemented');
end;


% ==== BACKUP DATA ====
if cfg.run, 
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir; 
        if any(datadir(end) == '\/'), datadir = datadir(1:end-1); end;
    end;
    % if default file name is used, generate a unique file name
    if strcmp(cfg.datafile, CFG.datafile),
        c = fix(clock);
        for i = 1:length(c),
            cfg.datafile = [cfg.datafile '-' num2str(c(i), '%02d')];
        end;
        if exist([CFG.datafile '.mat'], 'file'), delete([CFG.datafile '.mat']); end;
    elseif ~strcmp(datadir, pwd)
        % need to save in a different directory, so delete anyway
        if exist([cfg.datafile '.mat'], 'file'), delete([cfg.datafile '.mat']); end;
    end;    % otherwise just overwrite
    cfg.datafile = [datadir '/' cfg.datafile];
    save(cfg.datafile);
    dispx(['Final data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;

% -----------------------------------------------
% set output
if cfg.run,       % output solution
    % varargout = {approx};
elseif cfg.replay,              % output history and possibly figure handles
    varargout = {hist, figh};
elseif cfg.evol || cfg.sol,               % output fig handles (replay takes precedence)
    varargout = {figh};
end;

end
% lspi() RETURNING varargout =================================================================
