function [avgreg, reg, info] = simpleregret(cfg)
% Evaluate the simple regret of a policy
%   [AVGREG, REG, INFO] = SIMPLEREGRET(CFG)
% Inputs:
%   CFG     - inputs and config, see below
% Outputs:
%   AVGREG  - the average regret across the set of states provided, see below
%   REG     - regret for every state provided, row vector
%   INFO    - additional info about the evaluation process
%
% Supported types of policies:
%   - generic, assumed to be function handle for function of the form u = h(x)
%   - planning algorithm, assumed to be function name for planning algorithm, see e.g. OPD, OPSS, OLOP

CFG.h = [];                 % policy
CFG.Qstar = [];             % optimal Q-function (or approximation thereof)
CFG.hstar = [];             % optimal policy (or approximation thereof)
CFG.X = [];                 % evaluate regret at these states
    % can be:
    % 1) flat array of states: size dim(x) x N, where N is the number of states to consider
    % 2) cell array of grids
    % 3) a string name, in which case 'model' must be set, and model.problem('X0') will be
    %   called with the string to obtain the set of states
CFG.problem = [];           % required if X is a string

% process config
cfg = parseconfig(cfg, CFG);

% easy-access variables
h = cfg.h; hstar = cfg.hstar; Qstar = cfg.Qstar; X = cfg.X;

% process state set
if ischar(X), X = feval(cfg.problem, 'X0', X); end;
if iscell(X), X = flat(X); end;

% main loop
N = size(X, 2);
reg = zeros(1, N);
for i = 1:N,
    % use policy to choose an action
    u = feval(h, X(:, i));    
    % find optimal action
    ustar = feval(hstar, X(:, i));
    % compute regret
    reg(i) = feval(Qstar, X(:, i), ustar) - feval(Qstar, X(:, i), u);
    if reg(i) < 0, warning('Regret is negative!'); end;
    
    % update history
    % note we can't preinitialize as we don't know the size/structure etc.
	U(:, i) = u;
    Ustar(:, i) = ustar;
end;

% average regret
avgreg = mean(reg);

% return additional info if requested to do so
if nargout >= 3,
    % chosen actions and optimal actions
    info.U = U;
    info.Ustar = Ustar;
end;

