%% Optimistic planning experiments for the inverted pendulum
function dc_soop(mode, expids, custombcfg, customacfg, customplanparams)

%% General config
% clear all; close all;

% batch experiment
BCFG = struct;
BCFG.alg = 'genmpc';
BCFG.verb = 5;
% BCFG.idfields = {'alg', 'problem', 'planalg', 'planparamsn'};
BCFG.startat = 1; BCFG.repeat = 1;
BCFG.skipcache = 1;         % we don't need to cache the results, they are easy to compute

% algorithm
ACFG = struct;
ACFG.problem = 'dc_problem';
% model
ACFG.model_params = {'rewtype=nlqr Q=[1,0;0,.001] R=.05', 'type=lin maxx=[pi;15*pi] maxu=10 Ts=0.01'};
ACFG.online = 1;
ACFG.gamma = 0.95;
ACFG.trials = 1;
ACFG.maxsteps = 100;
ACFG.reset = [-pi, 0]';
% visualization
ACFG.verb = 3;
ACFG.datadir = pwd; %'c:\mfiles\approxrl\systems\invpend_setup\soop_ip';
ACFG.online = 1;

cd(ACFG.datadir);

% default custom configs (empty)
if nargin < 3, custombcfg = struct; end;
if nargin < 4, customacfg = struct; end;
if nargin < 4, customplanparams = struct; end;

    % nested utility function for executing experiment
    function execbatchexp(expid)
        if any(strcmp(expid, expids)),  % if experiment is requested
            bcfg = copyfields(custombcfg, bcfg); 
            acfg = copyfields(customacfg, acfg);
            acfg.planparams = copyfields(customplanparams, acfg.planparams);
            % check if we need to call 'process' or 'eval'
            if any(strcmp(mode, 'run')), 
                processbatch(bcfg, acfg); 
            elseif any(strcmp(mode, {'regret', 'return', 'exectime', 'horizon'})), 
                bcfg.out = mode;
                evalbatch(bcfg, acfg);
            end;
        end;
    end

%% Optimal solution
if any(strcmp('nearopt', expids)),
    model = dc_problem('model', ACFG.model_params{:});
    cfg.gamma = ACFG.gamma;
    cfg.model = model;
    [K, X, cfg] = gammadare(cfg);
%     pcfg = ccfg;         % copy relevant options
%     pcfg.datasource = 'caller';
%     pcfg.genh = 1;
%     pcfg.policy = @lsf;
%     pcfg.policyargs = {K, model.maxu};
%     pcfg.colorbar = 1;
%     pcfg.figname = sprintf('K=[%f, %f]', K(1), K(2));
%     cfg.verb = 0;
%     dc_plot(pcfg); 

    % traj
    cfg = struct;
    cfg.gamma = ACFG.gamma;
    cfg.policy = 'lsf';
    cfg.policyargs = {K, model.maxu};
    cfg.verb = 0;
    cfg.x0 = [-pi; 0];
    cfg.tend = model.Ts * ACFG.maxsteps;
    hist = clreplay(model, cfg);
    Rnearopt = hist.R(1)
end;

%% SOOP // CRAN Desktop
% Exper also available for alpha = 1 (i.e. always expand deepest) but only up to 10K, 15K too comp
% intensive (30min per step)

bcfg = BCFG;
bcfg.mode = 'run';
bcfg.idinfix = 'soop';
bcfg.legendprefix = 'SOOP';
bcfg.in = 'planparamsn';
bcfg.idfields = {'alg', 'problem', 'planparamsn', 'planparamsdimselect', 'planparamsalpha'};

acfg = ACFG;
acfg.planalg = 'sooplp';
acfg.planparams.method = 'soop';
acfg.planparams.n = [100 500 1000 2500 5000];
acfg.planparams.dimselect = 'disc';
acfg.planparams.alpha = 0.1:0.1:0.9;

execbatchexp(bcfg.idinfix);

%% LP // CRAN Desktop

bcfg = BCFG;
bcfg.mode = 'run';
bcfg.idinfix = 'lp';
bcfg.legendprefix = 'LP';
bcfg.in = 'planparamsn';
bcfg.idfields = {'alg', 'problem', 'planparamsn', 'planparamslipschitz'};

acfg = ACFG;
acfg.planalg = 'sooplp';
acfg.planparams.method = 'lp';
acfg.planparams.dimselect = 'lp';
acfg.planparams.n = [100 500 1000 2500 5000];
% very large execution times ... not feasible - investigate why
acfg.planparams.lipschitz = 0.1:0.1:1.5;    

execbatchexp(bcfg.idinfix);


%% HOLOP // CRAN Desktop
% 5 experiments available with detact=1

bcfg = BCFG;
bcfg.mode = 'run';
bcfg.idinfix = 'holop';
bcfg.legendprefix = 'HOLOP';
bcfg.in = 'planparamsn';
bcfg.startat = 1;
bcfg.repeat = 10;
bcfg.idfields = {'alg', 'problem', 'planparamsn', 'planparamsK', 'planparamsdetact', 'rep'};

acfg = ACFG;
acfg.planalg = 'holop';
acfg.planparams.n = [100 500 1000 2500 5000];
acfg.planparams.K = [5 10 15 20 25 30 40 50 75 100];
acfg.planparams.detact = 0;

execbatchexp(bcfg.idinfix);

%% OPD, discrete-action baseline // CRAN Desktop

bcfg = BCFG;
bcfg.mode = 'run';
bcfg.idinfix = 'opd';
bcfg.legendprefix = 'OPD';
bcfg.in = 'planparamsntrans';
bcfg.idfields = {'alg', 'problem', 'planparamsntrans', 'planparamsU'};

acfg = ACFG;
acfg.planalg = 'opd';
acfg.planparams.ntrans = [100 500 1000 2500 5000];
% form a set of equidistant action discretizations
M = 3:2:15;
U = cell(length(M), 1); for i = 1:length(M), U{i} = symequidgrid(M(i), 3); end;
acfg.planparams.U = U;

execbatchexp(bcfg.idinfix);

end     % main function