% 2-D Q learning for DC motor, modified by L. Busoniu
% on the basis of a file by Thijs Raamakers

clear all

% -------------------
% Create model
Ts = 0.05;

UseSetup = 0;
if UseSetup, initDCmotor; end

if ~UseSetup,
    %2nd order model
    tau = 0.106;        %R*J/(R*b+K^2)
    km = 18.4;
    model.A = [0 1; 0 -1/tau];
    model.B = [0; km/tau];
    model.C = eye(2);
    model.D = 0;

    %2nd order discrete-time state-space model; state = [theta dtheta]
    model.G = ss(model.A,model.B,model.C,model.D); 
    model.G = c2d(model.G,Ts,'zoh');
end

% -------------------
% Init learning

% cost function
gamma = 0.98;
Qlqr = diag([1, 0.0001]);
Rlqr = 0;

% compute LQR linear feedback
dcfg.gamma = gamma;
dcfg.A = model.G.a;
dcfg.B = model.G.b;
dcfg.Q = Qlqr;
dcfg.R = Rlqr;
dcfg.Ts = Ts;
Klqr = gammadare(dcfg);

% max values
xmax = [pi; 16*pi];
zmax = abs(Klqr * xmax);
umax = 10;
% create grids
Nhalf = [12 5]; ls = {logspace(-2 ,0, Nhalf(1)), logspace(-2 ,0, Nhalf(2))};
xgrid = {sortx([-xmax(1)*ls{1} 0 xmax(1)*ls{1}]), sortx([-xmax(2)*ls{2} 0 xmax(2)*ls{2}])};
Mhalf = 5; ls = logspace(-2 ,0, Mhalf);
ugrid = sortx([-umax*ls 0 umax*ls]);
% ugrid = [-10 -1 -0.5 -.2 -.1 0 .1 .2 0.5 1 10];
N = [length(xgrid{1}) - 1, length(xgrid{2}) - 1]; M = length(ugrid);

% Learning parameters
alphasched = 0;
if alphasched,
    alphaN = 1; alphaM = 1;
else
    alpha = 0.1;
end;
epsilon = 0.6; depsilon = .98;
lambda = 0.5;


% Number of episodes and steps
Ne = 700;
K = 5/Ts;

% -------------------
% init history
% history across episodes
X = cell(1, Ne); Z = cell(1, Ne);
U = cell(1, Ne); 
R = cell(1, Ne);
H = cell(1, Ne);    % policy storage
RC = nan(1, Ne);    % cumulative reward

% prepare figure
figh = figure; set(gcf, 'Position', [0 0 1000 800]); movegui(gcf, 'center');

% -------------------
% Init & run learning
Q = zeros([N, M]);
C = zeros([N, M]);    % visited counters
E = zeros([N, M]);    % elig trace -- for later use

% time vector
t = Ts*(0:K);

% Episodes loop
for e = 1:Ne,
    % history within an episode
    x = nan(2, K+1); z = nan(1, K);
    u = nan(1, K);
    r = nan(1, K+1);
    
    E(:) = 0;       % reset elig trace

    % random initial state
    if UseSetup,
        urandom = (2*rand-1)*umax;
        tic
        while toc < 0.1 
            dcmotor('Write',H,0,1,urandom,0); %give random control input to change position randomly
        end
        dcmotor('Write',H,0,1,0.0,0); %stop motor
        pause(1);   
        D = dcmotor('Read',H); %read initial position (and velocity)
        x(:, 1) = [D(3); 0];
    else
        x(:, 1) = (2 * rand(2, 1) - 1) .* xmax;
    end

    z(1) = Klqr * x(:, 1);
    xi = findbox(xgrid, x(:, 1));
    
    % steps loop
    for k = 1:K,

        % choose greedy action
        ubest = ugrid(Q(xi(1), xi(2), :) == max(Q(xi(1), xi(2), :)));    % set of actions for which Q is max
        u(k) = ubest(find(abs(ubest) == min(abs(ubest)), 1, 'first')); % smallest in magnitude
        ui = find(ugrid == u(k), 1, 'first');
        if rand < epsilon,      % exploratory action
            uie = ceil(rand * M);
            % if exploration changed the action, reset elig trace
            if uie ~= ui, E(:) = 0; end;
            ui = uie;
            u(k) = ugrid(ui);
        end
        
        % take action, observe next state
        if UseSetup,
           tic
           while toc < Ts,
               dcmotor('Write',H,0,1,u(k),0); 
           end
           D = dcmotor('Read',H);
           x(1, k+1) = D(3); 
           x(2, k+1) = (x(1, k+1) - x(1, k))/Ts;
        else
           x(:, k+1) = model.G.a * x(:, k) + model.G.b * u(k);
        end

        % box index of new state
        xiplus = findbox(xgrid, x(:, k+1));

        % reduced state
        z(k+1) = Klqr * x(:, k+1);
       
        % compute reward
        r(k+1) = -x(:, k)' * Qlqr * x(:, k) - u(k) * Rlqr * u(k);
       
        % replace eligibility of crt x-u pair
        E(xi(1), xi(2), ui) = 1;  
        % compute temporal difference
        delta = r(k+1) + gamma * max(Q(xiplus(1), xiplus(2), :)) - Q(xi(1), xi(2), ui);
        % compute visit-dependent alpha if set
        if alphasched,
            alpha = alphaN ./ (alphaM + C);
        else    % nothing to do, alpha is set to a constant
        end;
        % update Q-table
        Q = Q + delta * (alpha .* E);
        % anneal elig
        E = lambda * E;
        % update visits counter
        C(xi(1), xi(2), ui) = C(xi(1), xi(2), ui) + 1;
        % new state index becomes old state index
        xi = xiplus;
    end;
    
    % store statistics
    X{e} = x;
    Z{e} = z;
    U{e} = u;
    R{e} = r;
    % compute cumulated reward per step
    RC(e) = sum(r(2:end)) / K;      % do not include the first NaN in the sum, or the result will be NaN
    
    % compute current policy
    h = nan(N + 1);
    for xi1 = 1:N(1),
        for xi2 = 1:N(2),
            ubest = ugrid(Q(xi1, xi2, :) == max(Q(xi1, xi2, :)));    % set of actions for which Q is max
            h(xi1, xi2) = ubest(find(abs(ubest) == min(abs(ubest)), 1, 'first')); 
        end;
    end;
    % store in statistics
    H{e} = h;
    
    % anneal epsilon
    epsilon = depsilon * epsilon;
    
    % plot result
    if ~mod(e, 10),
        figure(figh);
        subplot(2,2,1); semilogy(1:e, RC(1:e)); ylabel('Reward per step');
        title(['Episode #' num2str(e)]);
        subplot(2,2,2); plot(t, x(1, :)); ylabel('\theta [rad]');
        subplot(2,2,3), plot(t, x(2, :)); ylabel('\omega [rad]');
        subplot(2,2,4); ph = pcolor(xgrid{1}, xgrid{2}, h'); caxis([-umax umax]); colormap gray;
        set(ph, 'EdgeColor', 'none');
        xlabel('x_1'); ylabel('x_2'); title('h(x)');
        drawnow;
    end;
end;


if UseSetup,
    dcmotor('Write', H, 0, 1, 0.0, 0);
    dcmotor('CloseAll');
end