function varargout = edro_rtproblem(mode, sys, varargin)
% Interact with the real-time EdRo System.
%   OUT = EDRO_RTPROBLEM(MODE, SYS, VARARGIN)
% This function conforms to the specifications established by SAMPLE_RTPROBLEM (TODO create this function)
%
% Calling modes:
% SYS = EDRO_RTPROBLEM('system', SYSCFG)
% Create system structure. 
%
% TODO document other modes

% Requirements on the call sequence and its timing:
%   - 'init' must be called before anything else
%
%   - 'reset' is called at the beginning of each new trajectory, 
%           immediately followed by 'writeu'
%   - thereafter, 'readx' immediately followed by 'writeu' are called exactly once every sample time
%
%   - 'close' is called before stopping the algorithm
%
% 'idle' should be called if the system should wait for longer than 1 sample time

% 2010-08-03: changed generation of initial states for 'randpos' so that a true uniform distribution
% on the safe interval is obtained (rather than bounding to safe interval, which resulted in a weird
% distribution)

% remember the old angle values, in order to estimate velocities
persistent angold uold ufilteredold

varargout = {};
switch mode,
    
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Time-critical operations (come first to avoid "case" overhead)
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
    % read angles, estimate angular velocities
    % this is a time-critical operation
    case 'readx',
        % read angles, pick up only those considered
        angall = edro_readall(sys);
        ang = angall(sys.xjoints);
        % form x = [column vector of angles; column vector of velocities]
        if ~isempty(angold),    x = [ang (ang - angold) ./ sys.Ts]; 
        else                    x = [ang 0*ang];
        end;
        angold = ang;
        % rearrange x so that we have angle, velocity, angle, velocity ...
        x = x'; varargout = {x(:)};
      
        
    % write actions to actuators
    % this is a time-critical operation
    case 'writeu',
        u = varargin{1};
        % saturate control action
        u = min(sys.maxu, max(-sys.maxu, u));
        % filter action (saturating also filtered), if using filter
        if ~isempty(sys.ufilt),
            ufiltered = min(sys.maxu, max(-sys.maxu, ...
                sys.ufilt.a .* uold - sys.ufilt.b .* ufilteredold));
            uold = u;
            ufilteredold = ufiltered;
            u = ufiltered;
        end;
        % initialize output vector for all actuators
        uall = zeros(5, 1); uall(sys.ujoints) = u;
        % read the angles for safety purposes etc.
        angall = edro_readall(sys);
        % if need to control other joints to keep them fixed, do that
        if ~isempty(sys.fixjoints),
            uall(sys.fixjoints) = sys.fixK .* (sys.fixref - angall(sys.fixjoints));
        end;
        % override whatever is outside of safety range, with safety controller
        unsafelo = angall < sys.angsafe(:, 1);
        unsafehi = angall > sys.angsafe(:, 2);
        if any(unsafelo),
            uall(unsafelo) = sys.ubounds(unsafelo);
            % disp(['ANGSAFE_low exceeded by ' num2str(unsafelo')]);
            % below is P controller
            % uall(unsafelo) = sys.safetyK(unsafelo) .* (sys.angsafe(unsafelo, 1) - angall(unsafelo));
        end;
        if any(unsafehi),
            uall(unsafehi) = -sys.ubounds(unsafehi);
            % disp(['ANGSAFE_hi exceeded by ' num2str(unsafehi')]);
            % uall(unsafehi) = sys.safetyK(unsafehi) .* (sys.angsafe(unsafehi, 2) - angall(unsafehi));
        end;
        % finally, make sure all commands are within the set bounds
        uall = min(sys.ubounds, max(-sys.ubounds, uall));
        % and write them
        edro_writeall(sys, uall);
        % output the actual command values of the considered joints
        % (they were already bounded above)
        varargout = {u};
                
    % set the system to idle
    case 'idle',
        % set all actuators to zero
        edro_writeall(sys, zeros(5, 1));
    
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Non-time-critical operations
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % create system structure
    case 'system',    
        % save sys config variable; 'sys' will get overwritten below
        if nargin >= 2, syscfg = sys; 
        else syscfg = struct;
        end;
        
        % Experimentally measured physical bounds and offsets, in radians
        % 1 = base, 2 = shoulder, 3 = elbow, 4 = hand, 5 = gripper rotation, 
        % 6 = gripper (not yet supported)
        % Joint 2:  [-0.8, 0.53] + = BACKWARD, - = FORWARD
        %           Offset: +0.08 => corrected interval [-0.72 0.61]
        %           Max velocity with umax=0.5: 1.2 going backward -1.3 going forward
        %           Safe region [-0.7, 0.45] + 0.08 = [-0.62, 0.53] (made stricter below)
        % Joint 3:  [-1.29, 1.81] + = UP, - = DOWN
        %           Offset: -0.25 => corrected interval [-1.54 1.56]
        %           Maximum velocity with umax=0.5: -3 going down 2.5 going up
        %           Safe region [-0.5, 1.1] -0.25 = [-0.75 0.85]
        %               - ONLY VALID if robot head has room to go below the table surface; otherwise
        %               will hit the table surface
        % Joint 4:  [-1.1, 1.4] + = UP, - = DOWN
        %           Offset: -0.1 => corrected interval [-1.2, 1.3]
        %           Max vel with umax=.5: -3 going down, 2.3 going up
        %           Safe region [-0.8, 1] -0.1 = [-0.9, 0.9]
        % Joint 5:  [-1.77, 1.77] + = LEFT, - = RIGHT
        %           Offset: -0.01 => negligible, taken to be 0
        %           Max vel with umax=.5: 2
        % Note max vels were observed with P controller, can be larger
        % Note some safety regions are shrunk below, to account for dynamics of both links 2 and 3
        % moving together
        
        % state bounds: top row -- angles; bottom row -- velocities
        % take the offsets into account when setting angle ranges
        XBOUNDS = [NaN,     0.55,       1,        1.2,    1.77;
                   NaN,     1.5,        3,          3,      2.5];
        % the values below do not take the safety ranges into account
        % XBOUNDS = [NaN,     0.61,       1.54,       1.2,    1.77;
        %            NaN,     1.5,        3,          3,      2.5];
        % safety bounds for the angles; note transpose
        ANGSAFE = [NaN,     -0.5,      -0.55,      -0.9,   -1.77;
                   NaN,     0.4,       0.75,       0.9,    1.77]'; 
        % angle offsets; should be added to raw angles given by sensors
        ANGOFFSETS = [0     0.08    -0.25   -0.1     0]; 
        % action bounds
        UBOUNDS = [NaN,     0.5,    0.5,    0.5,    0.5];
        % default weights on the states and actions in the quadratic reward 
        XREW    = [NaN,     1,      1,      1,      1;
                   NaN,     0.01,   0.01,   0.01,   0.01];
        UREW    = [NaN,     0.1,    0.1,    0.1,    0.1];
        
        % 1) fixed fields, e.g., given by the hardware, safety controllers
        sys = struct;
        sys.id = 'edro'; 
        sys.problem = @edro_rtproblem;
        sys.det = [];   % not sure yet if it's deterministic or not (there will be noise...)
        sys.fun = @edro_evaltrans;
        sys.xbounds = XBOUNDS(:);           % remember all bounds
        sys.angsafe = ANGSAFE;
        sys.angoffsets = ANGOFFSETS(:);
        sys.ubounds = UBOUNDS(:);
        % sys.safetyK = [3 3 3 3 3]';       % gain for safety controller; currently applying +/- maxu
        sys.len = [0.041 0.25 0.25 0.13];   % joint lengths: 2 offset, 2, 3, 4+5 (since gripper rotates on diff axis)

        % 2) configurable fields and options: SYSCFG contains the defaults
        % hardware config, joint selection, fixed-joint controllers
        SYSCFG.port = 'COM8';  % which port to use
        SYSCFG.xjoints = 5;    % select which joints to take into account into state signal (vector)
        SYSCFG.ujoints = [];   % select which joints to control (vector); default = same as xjoints
        SYSCFG.fixjoints = []; % select which joints to keep fixed (using proportional control)
        SYSCFG.fixref = [];    % around which references (defaults to zeros)
        SYSCFG.fixK = 3;       % proportional gain (can be column, per-joint)
        SYSCFG.Ts = .05;       % sampling time
        SYSCFG.ufilt = [];     % action filter: scalar = time constant of 1st order filter
        % reward config
        SYSCFG.rewtype = 'lqr';% reward type: 'lqr', 'vectorfield'
        % params for 'lqr', only diagonal weighting supported
        SYSCFG.qrew = [];      % state weights in the reward; by default, populated from XREW
        SYSCFG.rrew = [];      % action weights in the reward; by default, populated from UREW
        % params for 'vectorfield' - following v.f. in cartesian position coordinates
        % this is for the end of joint 3 (so the final link and the gripper are not considered)
        SYSCFG.vectorfield = @(t, pos, c, rad, vel, slope)...
            -4 * slope * ((pos-c)'*(pos-c) - rad^2) * (pos-c) + vel*[-pos(2)+c(2); pos(1)-c(1)];
        % default parameters for vector field
        % center, radius, multiplier for rotation-inducing factor in vector field, 
        % multiplier for "slope" towards circle region
        SYSCFG.vectorfield_params = {[.3; .25], .075, 5, 5};
        % process the config, copy it to the system structure
        syscfg = parseconfig(syscfg, SYSCFG);
        sys = copyfields(syscfg, sys);
        
        % 3) derived fields
        % process selection of state and control variables
        if isempty(sys.ujoints), sys.ujoints = sys.xjoints; end;
        sys.p = 2*length(sys.xjoints);
        sys.q = length(sys.ujoints);
        % bounds (in column vectors)
        sys.maxx = XBOUNDS(:, sys.xjoints); 
        sys.maxx = sys.maxx(:);     % results in the order: (angle, angular velocity, angle, ...)
        sys.maxu = UBOUNDS(sys.ujoints)';
       
        % if fixrefs empty, initialize to zero
        if isempty(sys.fixref), sys.fixref = zeros(length(sys.fixjoints), 1); end;
        
        % compute filter
        if ~isempty(sys.ufilt) && isscalar(sys.ufilt),
            % assumed time constant of first-order filter is given
            ufilttf = c2d(tf(1, [sys.ufilt/4 1]), sys.Ts);
            ufilt.tau = sys.ufilt;
            ufilt.a = ufilttf.num{1}(2);
            ufilt.b = ufilttf.den{1}(2);
            sys.ufilt = ufilt;
        end;
        
        % reward config
        switch sys.rewtype,
            case 'lqr',
                if isempty(sys.qrew),
                    sys.qrew = XREW(:, sys.xjoints);
                    sys.qrew = sys.qrew(:)';
                end;
                if isempty(sys.rrew),
                    sys.rrew = UREW(sys.ujoints);
                end;
            case 'vectorfield',
                % do nothing
            otherwise,
                error('Reward type [%s] unknown', sys.rewtype);
        end;
        
        % correctness checks
        if isempty(sys.xjoints) || isempty(sys.ujoints),
            error('At least one joint must be measured and controlled');
        elseif any(sys.xjoints == 6) || any(sys.ujoints == 6),
            error('Gripper closure and opening not supported');
        end;
        if any(sys.fixref < sys.angsafe(sys.fixjoints, 1)) ...
                || any(sys.fixref > sys.angsafe(sys.fixjoints, 2)),
            error('Fixed-joint references are not within safety region');
        end;
        
        % prepare output
        varargout = {sys};
        
    % initialize hardware system
    case 'init',
        % get offsets and gains for i/o converters
        % note we only need the values for the states that matter
        hwinit;
        sys.adinoffs = adinoffs(1:6);
        sys.adingain = adingain(1:6);
        sys.daoutoffs = daoutoffs(1:6);
        sys.daoutgain = daoutgain(1:6); 

        % com port config
        com = rs232('GetParams','default');             % get default RS232 parameters
        com.Port = sys.port;                            % define port number
        com.BaudRate = 115200;                          % define Baud rate
        com.ReadTimeout = 1;                            % define short timeout
        com.WriteTimeout = 1;                           % define short timeout
        [params,result] = rs232('open', com);           % open the port
        
        % place com object on system structure
        sys.com = com;
        
        % reset persistent variables, signaling they are "empty"
        angold = []; uold = 0; ufilteredold = 0;
        
        varargout = {sys};
       
        
    %  finalize and close connection with hardware
    case 'close', 
        % reset actuators, close the port
        rs232('write', sys.com, uint8(56));             % laser off
        rs232('write', sys.com, uint8(0));              % control byte for send all actuators
        rs232('write', sys.com, uint8(127*ones(6,1)));  % reset all 6 actuators to 0 (127 bin)
        rs232('close', sys.com);                        % close serial port
        % output system
        varargout = {sys};
        
        
    % reset system to an initial state
    case 'reset',
        reset = varargin{1};
        % support calling using only reset type
        if ischar(reset) && any(strcmp(reset, {'randpos'})),
            reset = struct('type', reset);
        end;
        
        % reset config
        RESET.type = 'randpos';
        reset = parseconfig(reset, RESET);
        
        switch reset.type,
            case 'randpos',
                % choose a random position (angle) WITHIN safe region for xjoints
                ref = sys.angsafe(sys.xjoints, 1) + ...
                    rand(length(sys.xjoints), 1) .* (sys.angsafe(sys.xjoints, 2) - sys.angsafe(sys.xjoints, 1));
                % stabilize xjoints AND fixed joints together
                % note that the stabilize function does NOT implement a safety controller
                x0 = edro_stabilize(sys, [sys.xjoints sys.fixjoints], [ref; sys.fixref]);
                % ignore any fixed joints when returning the state
                x0 = x0(1:2*length(sys.xjoints));
            otherwise,
                error('Reset type [%s] unknown', reset.type);
        end;

        % save the angle value
        angold = x0(1:2:sys.p);
        % reset control history
        uold = 0; ufilteredold = 0;

        varargout = {x0};
        
end;    % main 'mode' switch

end


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Evaluate edro transition
% Compute reward and terminal flag
function [rplus, term] = edro_evaltrans(sys, x, u, xplus)
    if sys.rewtype(1) == 'l',       % LQR
        rplus = - sys.qrew * x.^2 - sys.rrew * u.^2;
        
    elseif sys.rewtype(1) == 'v';   % vectorfield
        % this type of reward enforces the end of joint 3 to follow a trajectory defined
        % by the vectorfield
        % Assumption: x(1) = theta2; x(3) = theta3;
        % Note: Y value includes the offset of joint 2 from the robot's base (but not the height
        % of this base from the table surface)
        len = sys.len;  % for easy-access
        pos = [         - len(2)*sin(x(1))  - len(3)*sin(x(1)+x(3)-pi/2); ...
                len(1)  + len(2)*cos(x(1))  + len(3)*cos(x(1)+x(3)-pi/2)];
        % next position
        posplus=[       - len(2)*sin(xplus(1))  - len(3)*sin(xplus(1)+xplus(3)-pi/2); ...
                len(1)  + len(2)*cos(xplus(1))  + len(3)*cos(xplus(1)+xplus(3)-pi/2)];
        % reference position is given by integrating the vector field
        % from pos, for one sample time
        [odet odey] = ode45(sys.vectorfield, [0 sys.Ts], pos, [], sys.vectorfield_params{:});
        refpos = odey(end, :)';
        % distance between posplus and reference position gives the penalty
        rplus = -norm(posplus - refpos)^2;

    else rplus = [];
    end;
    term = 0;   % task is continuing (not episodic)
end

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Read all angles
% Output column vector ang
function ang = edro_readall(sys)
    % control byte for read D/A
    rs232('write', sys.com, uint8(128));
    % read 14 bytes from 7 sensors; place in matrix where each column is one sensor
    adin = reshape(rs232('read', sys.com, 14), 2, 7);
    % transform to double using big endian, only for the angles that matter
    % apply sensor gain and offset
    ang = (sys.adingain .* ([256 1] * double(adin(:, 1:6)) + sys.adinoffs))';
    % only return the first 5 angles, ignoring the gripper
    ang = ang(1:5) + sys.angoffsets;
end

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Write all actuators
% Input column vector u of 5 actuators
function edro_writeall(sys, u)
    % add a 0 for the gripper actuator
    % scale u values to digital range, convert to uint8, write
    daout = sys.daoutgain .* [u' 0] + sys.daoutoffs;   
    rs232('write', sys.com, uint8(0));              % control byte for send all actuators
    rs232('write', sys.com, uint8(daout));          % write 6 bytes to 6 actuators
end


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Stabilize robot joints
%   sys     - system structure
%   joints  - which joints to stabilize
%   ref     - around which reference
%   K gain, maxsteps, tol, tolsteps see code
function [xfinal, X] = edro_stabilize(sys, joints, ref, K, maxsteps, tol, tolsteps)
    % defaults
    if nargin < 4, K = 3; end;                  % proportional gain
    if nargin < 5, maxsteps = 50; end;          % maximum # of steps to control
    if nargin < 6, tol = 0.2; end;              % tolerance for controlled angles
    if nargin < 7, tolsteps = 4; end;           % if we are within tolerance for this # steps, stop

    % angle and angular velocity indices
    p = length(joints)*2;
    angi = 1:2:p-1;
    veli = 2:2:p;
    % initialize history
    X = zeros(p, maxsteps);

    % control loop
    k = 1; stop = 0;
    tic;
    while k <= maxsteps && ~stop,
        % read angles to control
        angall = edro_readall(sys); 
        ang = angall(joints);
        % compute and write controls
        uall = zeros(5, 1); 
        uall(joints) = max(-sys.ubounds(joints), min(sys.ubounds(joints), K .* (ref - ang)));
        edro_writeall(sys, uall);
        % save on history, estimate velocity if we have at least an earlier sample
        X(angi, k) = ang;
        if k > 1, X(veli, k) = (ang - X(angi, k-1)) ./ sys.Ts; end;
        % stop if we are within tolerance for tolsteps steps
        stop = k >= tolsteps && ...    
                all(all(abs(X(angi, k-tolsteps+1:k) - repmat(ref, 1, tolsteps)) < tol));
        % increment sample
        k = k + 1;
        % hold on for one sample
        while toc < sys.Ts, end; tic;
    end;
    
    % read angles one final time (after the last sample)
    angall = edro_readall(sys); 
    ang = angall(joints);
    X(angi, k) = ang;
    X(veli, k) = (ang - X(angi, k-1)) ./ sys.Ts;
    
    xfinal = X(:, k);
end
