function rtrobot_init(Port,ref,Kp)
% initialize the Ed-Ro robot to a desired pose (0 by default)
%    RTROBOT_INIT(Port,ref,Kp)
%
%       Port ... port, e.g., 'COM1' (default Port = 'COM1')
%       ref  ... reference to initialize to (default ref = 0)
%       Kp   ... proportional gain (default Kp = 3)

% (c) Robert Babuska, 2005

if nargin < 1, Port = 'COM1'; end;              % default port
if nargin < 2, ref = 0; end;                    % default reference
if nargin < 3, Kp = 5; end;                     % default control gain

h = 0.05;                                       % sampling period
eps = 0.05;                                     % tolerance around ref
umax = 0.5;                                     % maximum allowed control 

% initialize, open the port, define timer
hwinit;                                         % get HW parameters
n = length(joint);                              % number of joints to be controlled
joint = joint(:)+(0:6:6*(n-1))';                % convert into sequential index
joint_select = zeros(6,n);                      % prepare zeros
joint_select(joint) = 1;                        % assing 1 at the right places

com = xserial('Port',Port,'Baudrate',115200);   % initialize COM port
fopen(com);                                     % open serial port
rtclear all;                                    % clear all RT timers
rtdef(1,'tick',h);                              % define a tick timer
rtstart(1);                                     % start the timer

% control loop
rtwr(1,'t',0);                                  % reset the tick variable
for i = 1 : 100,                                % run for at most 100*h seconds
    while rtrd(1,'t') < i, end;                 % wait for a tick
    fwrite(com,128,'uint8');                    % control byte for read D/A
    x = double(fread(com,7,'uint16','b'))';     % read 14 bytes from 7 sensors
    x = adingain(:).*(x + adinoffs(:));         % scale to physical range
    x = joint_select'*x(1:6);                   % select the right variable
    if all(abs(ref-x) < eps); break; end;       % if sufficiently close, stop
    u = min(umax,max(-umax,Kp.*(ref-x)));       % saturate control action
    u = joint_select*u;                         % complete control vector
    u = daoutgain(1:6).*u' + daoutoffs(1:6);    % scale to digital range
    fwrite(com,0,'uint8');                      % control byte for send all actuators
    fwrite(com,u,'uint8');                      % write 6 bytes to 6 actuators
end;

% reset actuators, close the port
fwrite(com,56,'uint8');                         % laser off
fwrite(com,0,'uint8');                          % control byte for send all actuators
fwrite(com,127*ones(6,1),'uint8');              % reset all 6 actuators to 0 (127 bin)
fclose(com);                                    % close serial port
pause(.2);                                      % wait a while