function out = hillcar_problem(what, varargin)
% Car on the hill problem setup.
%   OUT = HILLCAR_PROBLEM(WHAT)
% Conforms to the specification of SAMPLE_PROBLEM.
% Special signatures:
%   MODEL = HILLCAR_PROBLEM('model', [CONFIG])
%       CONFIG      - model config, including numerical integration configuration
%   CFG = HILLCAR_PROBLEM('ce', [X0TYPE])
%       X0TYPE      - in ce mode, specify type of interesting state (see code)

% 2011-07-06: Changed second argument from only "DISC" discretization config to full model config

maxx = [1; 3];
maxu = 4;
xgrids = {-maxx(1):.05:maxx(1), -maxx(2):.2:maxx(2)};
ugrids = {[-maxu maxu]};
% default state for replays
x0 = (2*rand(2, 1)) .* maxx - maxx;       % a random pos-velocity
% x0 = [(2*rand-1) * maxx(1); 0];           % random position
% x0 = [0; 0];                              % zero state
gamma = 0.95;

switch what
    case 'info',    % basic info about the model (without creating the actual model)
        info.id = 'hillcar'; info.problem = @hillcar_problem;
        info.det = 1; info.p = 2; info.q = 1;
        out = info;
        
    case 'shortid',
        out = 'hc';
        
    case 'model',        
        % "set-in-stone" parameters: dimensionality etc.
        model.id = 'hillcar'; model.problem = @hillcar_problem;
        model.det = 1; model.p = 2; model.q = 1;
        
        % model configuration
        PAR.maxx = maxx; PAR.maxu = maxu; 
        % stochasticity
        PAR.stoch_prob = [];                   % if stochastic: probabilities with which...
        PAR.stoch_gain = [];                   % the action variable is multiplied with these values
        % reward
        PAR.rewtype = 'term';                  % 'term' terminal states only, -1 for failure; 
            % 'prog' progress indicator, varies linearly between 0 and progrew when position varies from
            % leftmost to rightmost; should be smaller than 1-gamma to avoid "drowning" terminal reward
            % also, failure reward is 0 not -1
        PAR.progrew = 0.01;                    % progress reward
        PAR.boundxterm = 1;                    % whether to bound terminal states to the domain
                % (default must be 1 for backward compatibility)
        % discretization options
        % note ode4 about twice as fast as ode45 but not so precise; still, quite close
        PAR.method = 'fixed-ode';              % 'ode', 'fixed-ode', 'euler'
        PAR.odesolver = 'ode4';                % solver function
        PAR.Ts = .1;                           % [sec] sample time
        PAR.odesteps = 10;                     % less accurate than 25 but faster
        if nargin >= 2,
            par = parseconfig(varargin{1}, PAR);
        else    % install defaults
            par = PAR;
        end;
        % split the discretization and other fields
        discfields = {'method', 'odesolver', 'Ts', 'odesteps'};
        disc = copyfields(par, struct, discfields);
        model = copyfields(par, model, setdiff(fieldnames(par), discfields));

        % process reward options
        model.maxr = 1;         % holds for all currently existing reward types
        
        % process discretization options
        switch(disc.method)
            case 'ode'
                disc.method = 1;
                disc.odet = [0 disc.Ts/2 disc.Ts];
                disc.odeopt = odeset;
            case 'fixed-ode'
                disc.method = 2;
                disc.odet = 0 : disc.Ts / disc.odesteps : disc.Ts;
                % the ode4-customized integration has been made obsolete
                % % use optimized MDP function when the integration method is ode4
                % and also optimized MC evaluation for fuzzy Qiteration
                % if strcmp(disc.odesolver, 'ode4'),
                %     model.fun = @hillcar_mdp_ode4;
                %     model.mc_fuzzyqiter_fun = @hillcar_mc_fuzzyqiter_ode4;
                % end;
            case 'euler'
                disc.method = 3;
        end;       
        model.disc = disc;
        model.Ts = disc.Ts;     % also required on model structure
 
        model.fun = @hillcar_mdp;
        model.plotfun = @hillcar_plot;
        
        % process stochastic options
        model.det = isempty(model.stoch_prob);
        if ~model.det, model = stochact_mdpwrapper(model, 'prepare'); end;
        
        out = model;
        
	case 'ce'
        % cfg.N = 10;       % can override default # of basis functions
        cfg.U = flat(ugrids);
        cfg.x0 = x0;
        cfg.gamma = gamma;
        if nargin >= 2, param = varargin{1};
        else            param = 'zero'; 
        end;
        cfg.X0 = hillcar_problem('X0', param);

        out = cfg;
        
	case 'X0'
        if nargin >= 2, param = varargin{1};
        else            param = 'zero'; 
        end;
        if ischar(param),
            switch param
                case 'zero',        % only zero state important
                    X0 = [0; 0];
                case 'pos',         % assume only initial speed 0 is important
                    X0 = flat({-1:.25:1, 0});
                case 'small',        % April 2010, small set containing pos and speed
                    X0 = flat({-1:.25:1, [-3 0 3]});
%                 case 'posspeed',    % OLD coarse grid in position and speed
%                     X0 = flat({-1:1/4:1, -3:3/4:3});
                case 'posspeed',        % July 2007, coarse grid in position and speed
                    X0 = flat({-1:.25:1, -3:1:3});
                case 'finest',        % fine grid in position and speed
                    X0 = flat({-1:.1:1, -3:.25:3});
                case 'finep',        % July 2007, medium grid in position and speed
                    X0 = flat({-1:.1:1, -3:1:3});
            end;
        else    % otherwise it is assumed param is already a grid of points
            X0 = param;
        end;

        out = X0;
        
    case 'reset',
        % note "sequence" not yet tested!
        RES.type = [];
        RES.x0 = [];
        RES.trial = [];
        model = varargin{1};
        res = varargin{2};
        if ischar(res) && any(strcmp(res, {'rand'})),   % handle simple method identifier
            res = struct('type', res);
        elseif isnumeric(res),
            res = struct('type', 'values', 'x0', res);
        end;
        
        res = parseconfig(res, RES);
        switch res.type,
            case 'rand',
                x0 = (2*rand(model.p, 1) - 1) .* model.maxx;
            case 'values',
                x0 = res.x0(:, unidrnd(size(res.x0, 2)));
            case 'seq',     % sequence of initial states, one per trial
                x0 = res.x0(:, mod(res.trial-1, size(res.x0, 2))+1);
        end;                    
        out = x0;
        
    case {'lspi', 'lspe'},
%         cfg.gamma = gamma;
        cfg.x0 = x0;
        out = cfg;
        
    case 'fuzzy'
        cfg.xgrids = xgrids;
        cfg.ugrids = ugrids;
        cfg.x0 = x0;
        cfg.gamma = gamma;
        out = cfg;

    case 'fittedqi';
        cfg.x0 = x0;
        out = cfg;        
        
end;


% END doubleint_problem(), RETURNING out ====================================