function rarm_lspi_old(exp, mode, c)
% Online and other LSPI experiments for robotic arm. This file handles the experiments used in the thesis
% chapter.
%
% For ICML paper draft and later, use the rarm_olspi file instead.
% 
% For all experiments except "r5new", uses an ad-hoc control flow with explicit code for calling 
% LSPI/LSPIONLINE (not via LSPIEXP) 
% LSPIEXP is only used to evaluate the performance
%
% Experiment settings include comments summarizing their results
% Also by checking datafile fields one should recover the meanings of the MAT files

% What experiments to process and in what mode (defaults)
if nargin < 1, exp = ''; end;
if nargin < 2, mode = 'run'; end;
if nargin < 3, c = struct; end;

% common config, including some fields specific to offline LSPI
CCFG = struct;
CCFG.(mode) = 1;
CCFG.problem = 'rarm_problem';
CCFG.datadir = pwd;
CCFG.gamma = 0.98;
CCFG.eps = 1e-3;
CCFG.maxiter = 50;
CCFG.sampledisp = .01;
CCFG.itersave = 10;
CCFG = setfigprop(CCFG, 'addfields');
ccfg = parseconfig(c, CCFG);

% common online config
OCFG = ccfg;
OCFG.alg = 'lspionline';
OCFG.eps = 0;
OCFG.maxtime = 60*60;   % run for a simulated hour
OCFG.trialtime = 10;    % seconds per trial
OCFG.dupd = 50 * 0.05;  % update each this # of samples
OCFG.explordecay = 0.001^(1/OCFG.maxtime);
OCFG.dstoretheta = OCFG.dupd;
OCFG.ddisp = 5;         % display stats after this #seconds
OCFG.dsave = 120;       % save once this #seconds
% output config
OCFG.startat = 1; OCFG.repeat = 5;
OCFG.show = {'rep'};
OCFG.donlineperf = [20:40:1000 1080:120:3600]; 
OCFG.X0 = 'pos';
OCFG.mc_eps = 0.1;
OCFG.plotminmax = 1;
OCFG.plotconf = 0;
OCFG.minmaxtype = 'line';
OCFG.xlim = [0 3600];
ocfg = parseconfig(c, OCFG);

% % common 
% ECFG = ocfg;
% ECFG.donlineperf = [20:40:1000 1080:120:3600]; 
% ECFG.X0 = 'pos';
% ECFG.mc_eps = 0.1;
% ECFG.plotminmax = 1;
% ECFG.plotconf = 0;
% ECFG.minmaxtype = 'line';
% ECFG.xlim = [0 3600];
% ecfg = parseconfig(c, ECFG);

% find baseline perf with fuzzyqi
%   Jbaseline = mc_fuzzyqi('datafile=rarm_fz_lqr X0=pos mc_eps=0.1');
%   result: Jbaseline = -216.1955 (quite larger than LSPI!!!)
% Interpolated:
%   Jbaseline = mc_fuzzyqi('datafile=rarm_fz_lqr X0=pos mc_eps=0.1 interph');
% result: Jbaseline = -217.1806

% % Code to compare LSPI performance with fuzzy QI performance:
% Using this code it appears that fuzzy QI has very bad performance for some states, and very
% good for others
% load d:\work\m\approxrl\data\lspipcache lspiorarmrbfdiscN5555M31
% lc = lspiorarmrbfdiscN5555M31;
% figure; mesh(lc.X0{1}, lc.X0{3}, squeeze(lc.Jtab)');
% [feJ, fJ] = mc_fuzzyqi('datafile=rarm_fz_lqr X0=pos mc_eps=0.1 interph');
% figure; mesh(lc.X0{1}, lc.X0{3}, squeeze(fJ)');

% ============================
% Online LSPI
% ============================

% 5x5x5x5 experiments. Ocassionally converge to quite good solutions, but most of the times
% don't exceed the quality of the solutions with 5x3^2
% These experiments were cleaned up with level 1.5
if checkrunexp('r5', exp), 
    switch mode,
        case 'run',
            lspiocfg = ocfg;            % start from the basic algorithm config
            lspiocfg.approx = 'enhanced type=rbfdisc N=[5,5,5,5] M=3 xspace=eq uspace=eq';
            lspiocfg.datafile = 'lspiorarm_rbfdiscN5555M3_#4'; % save to this data file
            lspiocfg.dsave = 120;
            lspionline(lspiocfg);          % call the function
        case 'onlineperf',
            % to evaluate the 5x5x5x5 experiments
            evalcfg = ocfg;
            % for these experiments the cached config has the default "lspi" algorithm
            % because it was not set to lspionline -- the algorithms were NOT run via lspiexp
            % so we use this hack to ensure the cache is properly re-used
                    evalcfg.alg = 'lspi';
            evalcfg.fixeddf = 'lspiorarm_rbfdiscN5555M3';
            evalcfg.startat = 1; evalcfg.repeat = 5; 
            lspiexp(evalcfg);
            % Note that the performance for run #1 is -114.1026 (by comparison with fuzzy QI)
        case 'stats',
            % to evaluate the 5x5x5x5 experiments
            evalcfg = ocfg;
            evalcfg.alg = 'lspionline';
            evalcfg.forceeval = 1;      % because not convergent
            evalcfg.fixeddf = 'lspiorarm_rbfdiscN5555M3';
            evalcfg.startat = 1; evalcfg.repeat = 1; 
            lspiexp(evalcfg);
    end;
end;


% ============================
% Offline LSPI
% ============================

% OFFLINE Experiments with 5^4 RBFs
% These experiments were cleaned up with level 2
if checkrunexp('roff5', exp), 
    switch mode,
        case {'run', 'resume'},
            lspicfg = ccfg;
            lspicfg.approx = 'enhanced type=rbfdisc N=[5,5,5,5] M=3 xspace=eq uspace=eq';
            lspicfg.samples = 'method=randdisc N=100000';
            lspicfg.datafile = 'lspirarm_rbfdiscN5555M3_randdiscNs100000'; 
            lspicfg.indexoptimized = 1;
            lspicfg.computecondA = 0;
            lspi(lspicfg);
    end;
end;

return;
% Experiments below this point were cleaned-up, archived, and removed

% % 5x5x5x5; the only difference from the r5 experiments is that these are run on the DCSC
% % desktop
% if checkrunexp('r5new', exp), 
%     lspiocfg = ocfg;
%     % these params are ignored top-level in LSPIEXP so we need to copy them into lspi_params
%     lspiocfg.type = 'rbfdisc';
%     lspiocfg.N = 5; lspiocfg.xspace = 'eq';
%     lspiocfg.M = 3; lspiocfg.uspace = 'eq';
%     lspiocfg.lspi_params = copyfields(lspiocfg, struct, {'maxtime', 'dstoretheta', 'ddisp', 'dsave'});
%     lspiexp(lspiocfg); 
% end;
    

% % RBF DISC, with 5x3 in every angle x velocity slice
% % These experiments were cleaned up with level 2
% if checkrunexp('r53', exp), 
%     switch mode,
%         case 'run',
%             % converges to something that makes sense, but it's very suboptimal, especially V in the 1st
%             % link slice; V is better in the 2nd link slice, though. By looking at the omegadot x alphadot
%             % slice, it appears that more RBFs would be needed. Policy slices are a mess...
%             lspiocfg = ocfg;            % start from the basic algorithm config
%             lspiocfg.run = 1;     
%             lspiocfg.approx = 'enhanced type=rbfdisc N=[5,3,5,3] M=3 xspace=eq uspace=eq';
%             lspiocfg.datafile = 'lspiorarm_rbfdiscN5353M3_#5'; % save to this data file (#1, #2, .., #5)
%             lspionline(lspiocfg);          % call the function
%         case 'onlineperf',
%             % to evaluate the 5x3x5x3 experiments
%             evalcfg = ocfg;
%             evalcfg.onlineperf = 1;
%             evalcfg.fixeddf = 'lspiorarm_rbfdiscN5353M3';
%             evalcfg.startat = 1;
%             evalcfg.repeat = 5;
%             evalcfg.donlineperf = [20:40:1000 1080:120:3600]; 
%             evalcfg.X0 = 'pos';
%             evalcfg.mc_eps = 0.1;
%             evalcfg.plotminmax = 1;
%             evalcfg.minmaxtype = 'line';
%             lspiexp(evalcfg);
%     end;
% end;

% % RBF DISC, with 3x3x3x3 RBFs
% % Trying with 3 RBFs only, since 5x3 gave similar results to 5x5 -- can't hurt
% % These experiments were cleaned up with level 2
% if checkrunexp('r3', exp), 
%     switch mode,
%         case 'run',
%             lspiocfg = ocfg;            % start from the basic algorithm config
%             lspiocfg.run = 1;     
%             lspiocfg.approx = 'enhanced type=rbfdisc N=[3,3,3,3] M=3 xspace=eq uspace=eq';
%             lspiocfg.datafile = 'lspiorarm_rbfdiscN3333M3_#5'; % save to this data file (#1, #2, etc)
%             lspionline(lspiocfg);          % call the function
%         case 'onlineperf',
%             % to evaluate the 3^4 experiments
%             evalcfg = ocfg;
%             evalcfg.onlineperf = 1;
%             evalcfg.fixeddf = 'lspiorarm_rbfdiscN3333M3';
%             evalcfg.startat = 1;
%             evalcfg.repeat = 5;
%             evalcfg.donlineperf = [20:40:1000 1080:120:3600]; 
%             evalcfg.X0 = 'pos';
%             evalcfg.mc_eps = 0.1;
%             evalcfg.plotminmax = 1;
%             evalcfg.minmaxtype = 'area';
%             lspiexp(evalcfg);
%     end;
% end;

% % OFFLINE Experiments with 5x3 RBFs
% % These experiments were cleaned up with level 2
% if checkrunexp('roff53', exp), 
%     switch mode,
%         case {'run', 'resume'},
%             % Does not converge
%             lspicfg = ccfg;
%             lspicfg.approx = 'enhanced type=rbfdisc N=[5,3,5,3] M=3 xspace=eq uspace=eq';
%             lspicfg.samples = 'method=randdisc N=120000';
%             lspicfg.run = 1;        
%             lspicfg.datafile = 'lspirarm_rbfdiscN5353M3_randdiscNs120000'; 
%             lspicfg.indexoptimized = 1;
%             lspicfg.computecondA = 0;
%             lspi(lspicfg);
%     end;
% end;


% % Does not converge
% lspicfg = ccfg;
% lspicfg.approx = 'enhanced type=rbfdisc N=[5,3,5,3] M=3 xspace=eq uspace=eq';
% lspicfg.samples = 'method=randdisc N=60000';
% lspicfg.run = 1;        
% lspicfg.datafile = 'lspirarm_rbfdiscN5353M3_randdiscNs60000'; 
% lspicfg.indexoptimized = 1;
% lspicfg.computecondA = 0;
% lspi(lspicfg);



% ============================
% ONLINE Experiments that did not save theta history
% ===========================

% % RBF DISC, trying 5x5x3x3
% % No theta h
% % Gives a poor solution (only 1 experiment ran)
% % It seems to struggle modeling the 2nd link
% lspiocfg = ocfg;            % start from the basic algorithm config
% lspiocfg.run = 1;     
% lspiocfg.approx = 'enhanced type=rbfdisc N=[5,5,3,3] M=3 xspace=eq uspace=eq';
% lspiocfg.datafile = 'lspiorarm_rbfdiscN5533M3';
% lspionline(lspiocfg);      

% % RBF + linear
% % The result is worse than using 5x3 RBFs + singleton (instead of linear)
% % Using more RBFs (antecedents) is basically impossible for computational reasons
% lspiocfg = ocfg;            % start from the basic algorithm config
% lspiocfg.run = 1;     
% lspiocfg.approx = 'enhanced type=rbflindisc N=[3,2,3,2] M=3 xspace=eq uspace=eq';
% lspiocfg.datafile = 'lspiorarm_rbflindiscN3232M3'; % save to this data file
% lspionline(lspiocfg);          % call the function

% % This was run on Zsofi's computer
% % Took an entire day, but best results so far
% % To evaluate final performance, use: (intermediate theta's not saved...)
% % lspiexp('problem=rarm_problem alg=lspionline stats forceeval fixeddf=lspiorarm_rbfdiscN5555M3_nothetah X0=pos mc_eps=0.1 show={}')
% % No theta h
% lspiocfg = ocfg;            % start from the basic algorithm config
% lspiocfg.run = 1;     
% lspiocfg.approx = 'enhanced type=rbfdisc N=[5,5,5,5] M=3 xspace=eq uspace=eq';
% lspiocfg.datafile = 'lspiorarm_rbfdiscN5555M3'; % save to this data file
% lspionline(lspiocfg);          % call the function

% % try with rarer updates of the parameter
% % RBF DISC, with 5x3 in every angle x velocity slice
% % Fails, reaches a solution that is not as good as that of the above
% % No theta h saved
% lspiocfg = ocfg;            % start from the basic algorithm config
% lspiocfg.run = 1;     
% lspiocfg.approx = 'enhanced type=rbfdisc N=[5,3,5,3] M=3 xspace=eq uspace=eq';
% lspiocfg.datafile = 'lspiorarm_rbfdiscN5353M3_Kupd100'; % save to this data file
% lspiocfg.dupd = 100 * 0.05;  % update each this # of samples
% lspionline(lspiocfg);          % call the function

