function varargout = rarm_visualize(cfg)
% Visualization function for DC motor and inverted pendulum

% REMARK 'step' (for online algs) visualization only supported independently from 'iter'
% TODO visualizing from data file: not tested
% output to sequence of figures: not tested

% default arguments
if nargin < 1, cfg = ''; end;

% where from to load the data, one of:
%       'caller'    - if the variables should be taken from the calling function
%       filename    - if data should be loaded from a file with name <filename>
% CFG.problem = 'cleanrob_problem';
CFG.datasource = 'caller';
CFG.gview = [];           % if nonempty: update existing; if empty: create new; graphical view
% CFG.datafile   = '';    % data source can be given as a data file here as well
% update/create view with:
CFG.mpcstep = 0;
CFG.finalize = 0;         % special option to finalize the output (namely, movie)
% which qiter/piter iteration/trial to plot
% CFG.ell = [];
% CFG.tau = [];
CFG.trial = [];
CFG.k = [];
CFG.ktotal = [];        % total #of steps from start of experiment

CFG = setfigprop(CFG, 'addfields');  % add figure settings fields to CFG
% output configuration
CFG.pause = .1;
CFG.plottarget = 'screen';      % 'latex', 'beamer', 'screen', 'movie'
CFG.snapshottarget = 'screen';  % 'latex', 'beamer', 'screen'
CFG.savedir = ''; 
CFG.savefig = 'rarmvisualization';% filename for saving figure
% if output is "movie":
CFG.compression = 'Cinepak';    % quality with Cinepak is pretty crappy...
CFG.quality = 100;
CFG.fps = 1;
CFG.snapshot = [];              % still snapshot config
% snapshot has the form [iter]

% process config
cfg = parseconfig(cfg, CFG);

% needed vars
vars = {'model', 'cfg'};
if cfg.mpcstep,
    vars = {vars{:}, 'X', 'REFX'}; 
end;

% retrieve needed variables from the data source
cfg1 = cfg;
switch cfg.datasource,
    case 'caller',              % load needed vars from caller space
        for i=1:length(vars), 
            eval(sprintf('%s=evalin(''caller'', ''%s'');', vars{i}, vars{i})); 
        end;
    otherwise,
%         load(cfg.datasource, vars{:});
end;
expcfg = cfg; cfg = cfg1;

% set up style...
sty = struct;
sty.cm = jet;
% sty.hcm = gray(128); sty.hcm = sty.hcm(33:end, :);    % w/o strong blacks and whites
sty.title = {'HorizontalAlignment', 'left', 'FontSize', 12, 'FontWeight', 'bold', 'Margin', 1, 'LineStyle', 'none'};
sty.linkcolors = {'b', 'r'};
sty.refbulletsize = 0.1;
sty.refcolor = 'k';

% ...and labels
% labels = struct;

% ===================================
% ==== Prepare data for plotting ====

% flags for what we are going to show
show = struct;
show.world = [];
show.x = [];
show.refx = [];
show.nothing = 0;

if cfg.mpcstep,
    show.world = 1;
    show.x = X(:, cfg.k+1, cfg.trial);
    if model.userefmodel, 
        % compute ref position indirectly by simulating the reference model (and ignoring its next
        % state)
        show.refpos = model.refmodel(model, REFX(:, cfg.k+1, cfg.trial), model.refmodel_params{:}); 
    end;
    algname = 'Online planning';
    gtitle = sprintf('%s, trial %d, time=%.1fs', algname, cfg.trial, cfg.ktotal*model.Ts);
    show.worldsubplot = {1, 1, 1};
else
    show.nothing = 1;       % signal nothing is in fact being shown
end;

% create or retrieve the view
if isempty(cfg.gview);  
    create = 1; gview = struct; gview.figh = figure; setfigprop(cfg);
    colormap(sty.cm); 
    gview.title = annotation('textbox', [.05 .93 .5 .05], 'String', gtitle, sty.title{:}) ;
    % open movie if needed
    if strcmp(cfg.plottarget, 'movie'),
        avifilename = [cfg.savedir cfg.savefig '.avi'];
        if exist(avifilename, 'file'), delete(avifilename); end;
        gview.aviobj = avifile(avifilename, 'compression', cfg.compression, 'quality', cfg.quality, 'fps', cfg.fps);
    end;
else
    create = 0; gview = cfg.gview; figure(gview.figh);
end;

% ===================================
% ==== Create world if needed =======
if ~isempty(show.world) && create,
    figure(gview.figh);
    subplot(show.worldsubplot{:});
    % draw circle and bullet representing current state
    gview.refpoint = rectangle('Position', [-1 -1 2 2] * sty.refbulletsize, 'Curvature', [1,1]);
    set(gview.refpoint, 'EdgeColor', 'none', 'FaceColor', sty.refcolor);
    gview.links(1) = line([0 model.len(1)], [0, 0], 'Color', sty.linkcolors{1}, 'LineWidth', 5);
    gview.links(2) = line(model.len(1)+[0 model.len(2)], [0, 0], 'Color', sty.linkcolors{2}, 'LineWidth', 5);
    workradius = 1.1*sum(model.len);
    xlim([-workradius workradius]); ylim([-workradius workradius]);
    axis square off manual
    drawnow;
end;

% ===================================================
% ==== Perform the update of the graphical view =====

if ~isempty(show.x),
    subplot(show.worldsubplot{:});
    pos0 = [0; 0];
    pos1 = [model.len(1)*cos(show.x(1)); model.len(1)*sin(show.x(1))];
    pos2 = pos1 + [model.len(2)*cos(show.x(1)+show.x(3)); model.len(2)*sin(show.x(1)+show.x(3))];
    set(gview.links(1), 'XData', [pos0(1) pos1(1)], 'YData', [pos0(2) pos1(2)]);
    set(gview.links(2), 'XData', [pos1(1) pos2(1)], 'YData', [pos1(2) pos2(2)]);
    set(gview.refpoint, 'Position', [show.refpos' 0 0] + [-1 -1 2 2] * sty.refbulletsize);
end;

% change the title if needed
if ~show.nothing,
    set(gview.title, 'String', gtitle);
end;

% ==========================================
% ==== Save graphical output (if any) ======

if ~show.nothing,
    switch cfg.plottarget,
        case 'screen',
            drawnow;
            if isscalar(cfg.pause), 
                if cfg.pause >= 0, pause(cfg.pause),
                else pause;
                end;
            else
                % assuming cell with structure 
                % {condition pause; condition pause; ...}; should include at
                % least a default where the condition evaluates true
                for i = 1:size(cfg.pause, 1),
                    if eval(cfg.pause{i, 1}),
                        if cfg.pause{i, 2} >= 0, pause(cfg.pause{i, 2}),
                        else pause;
                        end;
                        break;
                    end;
                    % otherwise try the next condition
                end;
            end;
        case 'movie',
            % add the figure as a frame to the movie
            drawnow;
            gview.aviobj = addframe(gview.aviobj, gview.figh);
            % determine if we need to take a snapshot at this stage
            if ~isempty(cfg.snapshot),
                if length(cfg.snapshot) == 1,
                    takesnapshot = ((cfg.fuzzyqiter || cfg.lspiter) && (cfg.ell == cfg.snapshot));
                elseif length(cfg.snapshot) == 2,
                    takesnapshot = ((cfg.lspionlinestep || cfg.lspihonlinestep || cfg.mpcstep) ...
                        && (cfg.trial == cfg.snapshot(1)) && (cfg.k == cfg.snapshot(2)));
                else
                    takesnapshot = 0;
                end;
            else
                takesnapshot = 0;
            end;
            if takesnapshot,
                % disp('snapshot');
                saveplot(gview.figh, [cfg.savedir cfg.savefig], cfg.snapshottarget);
            end;
    end;
end;

% Finalize movie if requested to do so
if cfg.finalize,
    if strcmp(cfg.plottarget, 'movie'),
        gview.aviobj = close(gview.aviobj);
    end;
end;

varargout = {gview.figh, gview};

end
% END =================================================

